module Main (main) where

import qualified LLVM.Extra.ScalarOrVector as SoV
import qualified LLVM.Extra.Arithmetic as A
import LLVM.Extra.Control (arrayLoop, )

import qualified LLVM.Core as LLVM
import LLVM.Core (Value, valueOf, Function, CodeGenModule, CodeGenFunction, )
import LLVM.ExecutionEngine (simpleFunction, )

import qualified Foreign.Storable as St
import Foreign.Marshal.Array (allocaArray, )
import Foreign.Ptr (Ptr, )

import qualified System.IO as IO

import Data.Word (Word32, )

import Control.Monad (liftM2, void, )

waveSaw :: Value Float -> CodeGenFunction r (Value Float)
waveSaw t =
   A.sub (valueOf 1) =<<
   A.mul (valueOf 2) t

osciSaw ::
   Value Float -> Value Float ->
   CodeGenFunction r (Value Float, Value Float)
osciSaw freq phase =
   liftM2 (,) (waveSaw phase) (SoV.incPhase freq phase)

waveSin :: Value Float -> CodeGenFunction r (Value Float)
waveSin t = A.sin t

moduleSaw :: CodeGenModule (Function (Word32 -> Float -> Ptr Float -> IO Float))
moduleSaw =
   LLVM.createFunction LLVM.ExternalLinkage $ \ size freq ptr -> do
      s <- arrayLoop size ptr (valueOf 0) $ \ ptri phase0 -> do
         (y, phase1) <- osciSaw freq phase0
         LLVM.store y ptri
         return phase1
      LLVM.ret (s :: Value Float)

writeSawCode :: IO ()
writeSawCode = do
   m <- LLVM.newModule
   void $ LLVM.defineModule m moduleSaw
   LLVM.writeBitcodeToFile "saw.bc" m

fillBuf f freq len h =
    allocaArray len $ \ ptr ->
        f (fromIntegral len) freq ptr >>
        IO.hPutBuf h ptr (len*St.sizeOf(undefined::Float))

renderSaw :: IO ()
renderSaw = do
   fill <- simpleFunction moduleSaw
   IO.withFile "saw.f32" IO.WriteMode $ \h ->
      let len = 10000
      in fillBuf fill 0.01 len h >>
         fillBuf fill 0.02 len h >>
         fillBuf fill 0.03 len h

main :: IO ()
main = do
   LLVM.initializeNativeTarget
   writeSawCode
   renderSaw
