module Main (main) where

import Data.List

import qualified LLVM.Extra.ScalarOrVector as SoV
import qualified LLVM.Extra.Arithmetic as A
import LLVM.Extra.Control (arrayLoop, )

import qualified LLVM.Core as LLVM
import LLVM.Core (Value, valueOf, Function, CodeGenModule, CodeGenFunction, )
import LLVM.ExecutionEngine (simpleFunction, )

import qualified Foreign.Storable as St
import Foreign.Marshal.Array (allocaArray, )
import Foreign.Ptr (Ptr, )

import qualified System.IO as IO

import Data.Word (Word32, )

import Control.Monad (liftM2, void, forM_)


waveSaw :: Value Float -> CodeGenFunction r (Value Float)
waveSaw t = do
   f <- A.mul (valueOf 2) t
   A.sub (valueOf 1) f
   
waveSine :: Value Float -> CodeGenFunction r (Value Float)
waveSine t = A.sin t

osciSaw ::
   Value Float -> Value Float ->
   CodeGenFunction r (Value Float, Value Float)
osciSaw freq phase =
   liftM2 (,) (waveSaw phase) (SoV.incPhase freq phase)


moduleSaw :: CodeGenModule (Function (Word32 -> Ptr Float -> Float -> IO Float))
moduleSaw =
   LLVM.createFunction LLVM.ExternalLinkage $ \ size ptr freq -> do
      s <- arrayLoop size ptr (valueOf 0) $ \ ptri phase0 -> do
         (y, phase1) <- osciSaw freq phase0
         LLVM.store y ptri
         return phase1
      LLVM.ret (s :: Value Float)

writeSawCode :: IO ()
writeSawCode = do
   m <- LLVM.newModule
   void $ LLVM.defineModule m moduleSaw
   LLVM.writeBitcodeToFile "saw.bc" m

step :: Float
step = 2**(1/12)
toneSaw :: Int -> Float
toneSaw idx = a * step**(realToFrac idx) / 44100.0 where
  a = 440

_a, _h, _c, _d, _e, _f, _g, _A :: Int
_a = 0
_h = 2
_c = 3
_d = 5
_e = 7
_f = 8
_g = 10
_A = 12

_1, _2, _4 :: Int
_1 = 44100
_2 = div _1 2
_4 = div _1 4
_pt = div _1 32


renderSaw :: IO ()
renderSaw = do
   fill <- simpleFunction moduleSaw
   IO.withFile "saw.f32" IO.WriteMode $ \h ->
      let (lens, freqs) = lied1
      in  forM_ (zip freqs lens) $ \(freq,len) -> allocaArray len $ \ ptr ->
             fill (fromIntegral len) ptr freq >>
             IO.hPutBuf h ptr (len*St.sizeOf(undefined::Float))

-- Alle meine Entchen
lied1 :: ([Int],[Float])
lied1 = (intersperse _pt [_2, _2, _2, _2, _1, _1, _2, _2, _2, _2, _1, _2, _2, _2, _2, _1, _2, _2, _2, _2, _1, _1, _2, _2, _2, _2, _1], intersperse 0 $ map toneSaw [_c, _d, _e, _f, _g, _g, _A, _A, _A, _A, _g, _A, _A, _A, _A, _g, _f, _f, _f, _f, _e, _e, _d, _d, _d, _d, _c])

-- fast Seven Nations Army
lied2 :: ([Int],[Float])
lied2 = (intersperse _pt $ [_1, _2, _4, _2, _2, _2, _2, _1]
        ,intersperse 0 $ map toneSaw [2, 2, 4, 2, 0, -2, -4])

main :: IO ()
main = do
   LLVM.initializeNativeTarget
   writeSawCode
   renderSaw
